from __future__ import annotations
from typing import Any, Dict, Iterable
from zormpg.base import DataModel
import peewee as pee
import json
import copy
import logging
logger = logging.getLogger(__name__)


class DictValue:

    def __init__(self, default) -> None:
        """
        default: default value
        the type of default value must be in str, int, float, bool, list, dict
        """
        self.default = default
        t = type(self.default)
        if t not in [str, int, float, bool, list, dict]:
            raise Exception(f'默认值类型必须在str, int, float, bool, list, dict中,但获得的是{t}')
        self.key = None
        self.owner: DictModel = None
        self._cache = None

    @property
    def cache(self):
        return copy.deepcopy(self._cache)

    @cache.setter
    def cache(self,value):
        self._cache = value

    def _get_row(self) -> DictModel:
        return self.owner.select().where(self.owner.__class__._model_key == self.key).get()

    def _get(self):
        if self.cache is not None:
            return self.cache
        v = self.default
        n = self._count()
        if n == 0:
            self.cache = self.default
            return self.cache
        try:
            v = self._get_row()._model_value
            self.cache = json.loads(v)
            return self.cache
        except:
            logger.warning(f'表{self.owner._table_name}的键{self.key}的值{v}解析失败.返回原始值')
            return v

    def _count(self):
        return self.owner.select().where(self.owner.__class__._model_key == self.key).count()

    def _set(self, value):
        if self._count() == 0:
            row = self.owner.__class__()
            row._model_key = self.key
        else:
            row = self._get_row()
        row._model_value = json.dumps(value)
        row.save()
        self.cache = value
        logger.debug(f'设置表 {self.owner._table_name} 的 {self.key} 为 {value}')


class DictModel(DataModel):

    _model_key = pee.CharField()
    _model_value = pee.TextField()

    def __init__(self, *args, **kwargs):
        self._stored_dict: Dict[str, DictValue] = {}
        for k, v in self.__class__.__dict__.items():
            if isinstance(v, DictValue):
                v.key = k
                v.owner = self
                self._stored_dict[k] = v
        super().__init__(*args, **kwargs)

    def __getattribute__(self, _name: str) -> Any:
        if _name == '_stored_dict':
            return super().__getattribute__(_name)
        if _name in self._stored_dict:
            return self._stored_dict[_name]._get()
        return super().__getattribute__(_name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == '_stored_dict':
            return super().__setattr__(__name, __value)
        if __name in self._stored_dict:
            return self._stored_dict[__name]._set(__value)
        else:
            return super().__setattr__(__name, __value)



import peewee as pee


class ZormPgConnect():

    @classmethod
    def db_config(cls,
                  db_name: str = None,
                  user: str = None,
                  password: str = None,
                  host: str = None,
                  port: int = 5432,
                  autorollback: bool = True,
                  ):
        cls.db = pee.PostgresqlDatabase(
            db_name,
            user=user,
            password=password,
            host=host,
            port=port,
            autorollback=autorollback
        )

    db = None



from setuptools import setup

setup(
    name='z-orm-pg',
    version='0.1.1',
    author='homo-zhou',
    author_email='408088242@qq.com',
    url='http://127.0.0.1',
    description='An orm for postgresql using peewee',
    packages=['zormpg'],
    install_requires=['psycopg2','peewee'],
    entry_points={
        'console_scripts': []
    }
)


from typing import Dict
import peewee as pee
from zormpg.connect import ZormPgConnect
from playhouse.migrate import PostgresqlMigrator, migrate

import logging
logger = logging.getLogger(__name__)

TYPE_MAP = {
    pee.CharField: "character varying",
    pee.IntegerField: "integer",
    pee.PrimaryKeyField: "integer",
    pee.BigIntegerField: "bigint",
    pee.TextField: "text",
    pee.DoubleField: "double precision",
    pee.ForeignKeyField: "integer"
}

DEFAULT_VALUE = {
    pee.CharField: "",
    pee.IntegerField: 0,
    pee.BigIntegerField: 0,
    pee.TextField: "",
    pee.DoubleField: 0
}

db = ZormPgConnect.db

class DataModel(pee.Model):

    _table_name:str = 'data_model'

    id = pee.PrimaryKeyField()

    class Meta:
        database = db
        def table_function(c): return c._table_name

    @classmethod
    def migrate(cls):
        if '_table_name' not in cls.__dict__:
            raise Exception('请用类静态属性:_table_name 指定表名')
        if cls.table_exists():
            logger.info(f"已存在表: {cls.__name__}")
            cls.migrate_columns()
        else:
            logger.info(f"创建表:   {cls.__name__}")
            db.create_tables([cls])

    @classmethod
    def get_fields(cls):
        fields: Dict[str, pee.Field] = {}
        for key, value in cls.__dict__.items():
            field = getattr(value, "field", False)
            if isinstance(field, pee.Field):
                fields[key] = field
        return fields

    @classmethod
    def migrate_columns(cls):
        fields = cls.get_fields()
        table = cls._table_name
        sql = """ SELECT column_name, data_type
            FROM information_schema.COLUMNS
            WHERE table_name = '%s'; """ % table
        existed_cols = dict(db.execute_sql(sql).fetchall())
        col_to_field = {}
        for field in fields.values():
            if isinstance(field, pee.ForeignKeyField):
                if field.model != cls:
                    continue
            if field.default == None:
                field.default = DEFAULT_VALUE.get(type(field),None)
            col_to_field[field.column_name] = field
        need_drop = []
        need_add = []
        for col, field in col_to_field.items():
            if col in existed_cols:
                if TYPE_MAP[type(field)] == existed_cols[col]:
                    continue
                need_drop.append(col)
            need_add.append(col)

        for col in existed_cols:
            if col not in col_to_field:
                need_drop.append(col)

        migrator = PostgresqlMigrator(db)
        operations = []
        for name in need_drop:
            operations.append(migrator.drop_column(table, name))
        for name in need_add:
            operations.append(migrator.add_column(table, name, col_to_field[name]))
        migrate(*operations)
